# -*- encoding: utf-8 -*-
__version__ = '0.0.6'
DEFAULT_PORT = 5672
AMQP_EXCHANGE = "amqppy"
AMQP_BROKER = "amqp://localhost:{}//".format(DEFAULT_PORT)

# how to publish a package to pypi
# https://packaging.python.org/distributing/

import logging
try:
    # not available in python 2.6
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):

        def emit(self, record):
            pass

# Add NullHandler to prevent logging warnings
logging.getLogger(__name__).addHandler(NullHandler())


class ResponseTimeout(Exception):
    pass


class PublishNotRouted(Exception):
    pass


class ExchangeNotFound(Exception):
    pass


class AbortConsume(Exception):
    pass


class DeadLetterMessage(Exception):
    pass



# -*- encoding: utf-8 -*-
import sys
import amqppy
import pika

if sys.version_info > (3,):
    import urllib.parse as urlparse  # pylint: disable=E0611,F0401
else:
    import urlparse


def create_url(host="localhost", port=5672, username="guest", password="guest", virtual_host="/", transport="amqp"):
    return "{}://{}:{}@{}:{}/{}".format(transport, username, password, host, port, virtual_host)


def parse_url(url):
    scheme = urlparse.urlparse(url).scheme
    schemeless = url[len(scheme) + 3:]
    # parse with HTTP URL semantics
    parts = urlparse.urlparse('http://' + schemeless)
    path = parts.path or ''
    path = path[1:] if path and path[0] == '/' else path
    return dict(
        transport=scheme,
        host=urlparse.unquote(parts.hostname or '') or None,
        port=parts.port or amqppy.DEFAULT_PORT,
        username=urlparse.unquote(parts.username or '') or None,
        password=urlparse.unquote(parts.password or '') or None,
        virtual_host=urlparse.unquote(path or '') or "/",
        **dict(urlparse.parse_qsl(parts.query)))


def _create_connection(broker, heartbeat_sec=None):
    # builds AMQP connection for each broker URL
    params = parse_url(broker)
    return pika.BlockingConnection(pika.ConnectionParameters(
        host=params["host"],
        port=int(params["port"]),
        virtual_host=params["virtual_host"],
        credentials=pika.PlainCredentials(params["username"], params["password"]),
        socket_timeout=1.25,
        heartbeat_interval=heartbeat_sec))



# -*- encoding: utf-8 -*-
import sys
import os
import pika
import uuid
import time
import logging
# add amqppy path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import amqppy
from amqppy import utils


logger = logging.getLogger(__name__)


def publish(broker, routing_key, body, headers=None, exchange=amqppy.AMQP_EXCHANGE, persistent=True):
    """Publish a message to the given exchange, routing key. This call creates a connection and once the message is sent the connection will be closed.
    Use class Publisher in case you want to reuse the same connection to send many messages.

    :param str broker: The URL for connection to RabbitMQ. Eg: 'amqp://serviceuser:password@rabbit.host:5672//'
    :param str rounting_key: The routing key to bind on
    :param str body: A json text is recommended. The body of the message you want to publish.
    :param dict headers: Message headers.
    :param str exchange: The exchange you want to publish the message.
    :param bool persistent: Makes message persistent. The message would not be lost after RabbitMQ restart.
    """
    publisher = Publisher(broker=broker)
    publisher.publish(exchange=exchange, routing_key=routing_key, body=body, persistent=persistent, headers=headers)


def rpc_request(broker, routing_key, body, exchange=amqppy.AMQP_EXCHANGE, timeout=10):
    """Makes a RPC request and returns its response. https://www.rabbitmq.com/tutorials/tutorial-six-python.html
    This call creates and destroys a connection every time, if you want to save connections, please use the class Rpc.

    :param str broker: The URL for connection to RabbitMQ. Eg: 'amqp://serviceuser:password@rabbit.host:5672//'
    :param str rounting_key: The routing key to bind on
    :param str body: A json text is recommended. The body of the request.
    :param str exchange: The exchange you want to publish the message.
    :param bool timeout: Maximum seconds to wait for the response.
    """
    rpc = Rpc(broker=broker)
    return rpc.request(exchange=exchange, routing_key=routing_key, body=body, timeout=timeout)

####################################################################


class Publisher(object):
    def __init__(self, broker):
        self._connection = utils._create_connection(broker=broker)

    def __del__(self):
        logger.debug("publisher destructor")
        if self._connection and self._connection.is_open:
            logger.debug("closeing connection")
            self._connection.close()

    def publish(self, exchange, routing_key, body, headers=None, persistent=True):
        """Publish a message to the given exchange, routing key.

        :param str exchange: The exchange you want to publish the message.
        :param str rounting_key: The rounting key to bind on
        :param str body: A json text is recommended. The Message you want to publish.
        :param dict headers: Message headers.
        :param bool persistent: Makes message persistent. The message would not be lost after RabbitMQ restart.
        """
        logger.debug("creating channel")
        channel = self._connection.channel()
        try:
            channel.confirm_delivery()
            logger.debug("publishing message at exchange: {} and routing_key: {}".format(exchange, routing_key))
            publish_result = channel.basic_publish(exchange=exchange,
                                                   routing_key=routing_key,
                                                   properties=pika.BasicProperties(
                                                       delivery_mode=2 if persistent else 1, headers=headers
                                                   ),  # 2 -> persistent
                                                   body=utils.json_dumps(body) if isinstance(body, dict) else body,
                                                   mandatory=True)  # to know if the message was routed
            if not publish_result:
                logger.debug("Publisher published message was not routed")
                raise amqppy.PublishNotRouted("Publisher published message was not routed")
        except pika.exceptions.ChannelClosed as e:
            if "NOT_FOUND - no exchange" in str(e):
                raise amqppy.ExchangeNotFound(str(e))

        finally:
            if channel and channel.is_open:
                logger.debug("closing channel")
                channel.close()


class Rpc(object):
    # def __init__(self, broker=None, host=None, port=5672, username="guest", password="guest", virtual_host="/"):
    #    self._connection = amqp_utils.__create_connection(broker=broker, host=host, port=port, username=username, password=password, virtual_host=virtual_host)

    def __init__(self, broker):
        self._connection = utils._create_connection(broker=broker)

    def __del__(self):
        logger.debug("rpc publisher destructor")
        if self._connection and self._connection.is_open:
            logger.debug("closeing connection")
            self._connection.close()

    def on_response(self, ch, method, props, body):
        if self.corr_id == props.correlation_id:
            logger.debug("on_response: {}".format(body))
            self.response = body

    def request(self, exchange, routing_key, body, timeout):
        """Makes a RPC request and returns its response. https://www.rabbitmq.com/tutorials/tutorial-six-python.html
        This call creates and destroys a connection every time, if you want to save connections, please use the class Rpc.

        :param str rounting_key: The routing key to bind on
        :param str body: A json text is recommended. The body of the request.
        :param str exchange: The exchange you want to publish the message.
        :param bool timeout: Maximum seconds to wait for the response.
        """
        self.exchange = exchange
        channel = self._connection.channel()
        # Enabled delivery confirmations:
        # very important to know if the message was delivered or consumed
        channel.confirm_delivery()

        try:
            # rpc response queue
            self.response_queue = channel.queue_declare(exclusive=True).method.queue
            channel.queue_bind(queue=self.response_queue, exchange=self.exchange, routing_key=self.response_queue)

            # lets listen response
            channel.basic_consume(queue=self.response_queue, consumer_callback=self.on_response, no_ack=True)

            logger.debug("publishing rpc request, exchange: {}, routing_key: {}, body: {}".format(self.exchange, routing_key, body))
            self.response = None
            self.corr_id = str(uuid.uuid4())
            publish_result = channel.basic_publish(exchange=self.exchange,
                                                   routing_key=routing_key,
                                                   properties=pika.BasicProperties(
                                                       reply_to=self.response_queue,
                                                       correlation_id=self.corr_id,
                                                       content_type='application/json',
                                                       delivery_mode=1),  # 2 -> persistent
                                                   body=body,
                                                   mandatory=True)
            if not publish_result:
                logger.debug("Rpc published message was not routed")
                raise amqppy.PublishNotRouted("Rpc published message was not routed")

            # wait for response
            logger.debug("waiting for rpc response... on \'{}\' for {} seconds".format(self.response_queue, timeout))
            start = time.time()
            while self.response is None:
                time.sleep(0.1)
                self._connection.process_data_events()
                if timeout > 0 and time.time() - start >= timeout:
                    logger.warning("AMQP RPC Timeout has been triggered waiting for the response")
                    raise amqppy.ResponseTimeout("AMQP RPC Timeout has been triggered waiting for the response")
            return self.response
        finally:
            if channel and channel.is_open:
                logger.debug("closing channel")
                channel.close()



# -*- encoding: utf-8 -*-
import traceback
from functools import wraps
import collections
# from collections import namedtupla
import pika
import logging
import threading
import time
import json
import sys
import os

# add amqppy path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import amqppy
from amqppy import utils


logger = logging.getLogger(__name__)

_ChannelExchange = collections.namedtuple('ChannelExchange', ['channel', 'exchange'])


class Worker(object):
    """ Encapsulate the worker setup.

    Stablish a connection to the AMQP broker, and opens a channel for each consumer task.
    conn = Connection(host="localhost", userid="guest",password="guest", virtual_host="/")
    """
    def __init__(self, broker, heartbeat_sec=None):
        self._conn = utils._create_connection(broker=broker, heartbeat_sec=heartbeat_sec)
        # map(callback) -> (channel, exchange)
        self._callbacks = {}
        self.quit = False
        self.thread = None

    def __del__(self):
        logger.debug("consumer worker destructor")
        self._close()

    def _close(self):
        for callback in self._callbacks:
            if self._callbacks[callback].channel and self._callbacks[callback].channel.is_open:
                self._callbacks[callback].channel.close()
        self._callbacks = {}

        if self._conn:
            logger.debug('closing connection')
            self._conn.close()
            self._conn = None

    def _create_channel(self, exchange, request_func):
        try:
            channel = self._conn.channel()
            channel.exchange_declare(exchange=exchange, exchange_type="topic", passive=True)
            self._callbacks[request_func] = _ChannelExchange(channel, exchange)
            return channel
        except Exception:
            channel = self._conn.channel()
            channel.exchange_declare(exchange=exchange, exchange_type="topic", passive=False, durable=True, auto_delete=True)
            self._callbacks[request_func] = _ChannelExchange(channel, exchange)
            return channel

    def stop(self):
        logger.debug("stop")
        self.quit = True
        self.join()
        self._close()

    def add_request(self, routing_key, request_func, exchange=amqppy.AMQP_EXCHANGE, durable=False, auto_delete=True,
                    exclusive=False):
        """ Register a new consumer task. These tasks will be execute when a new message arrive
        at :param:`queue_name`, which is binded to the exchange with :param:`routing_key`.
        """
        logger.debug("adding request, exchange: {}, topic: {} --> {}".format(exchange, routing_key, request_func))
        channel = self._create_channel(exchange, request_func)
        channel.queue_declare(queue=routing_key, durable=durable, auto_delete=auto_delete)
        channel.queue_bind(queue=routing_key, exchange=exchange, routing_key=routing_key)
        channel.confirm_delivery()
        channel.basic_consume(
            exclusive=exclusive,
            queue=routing_key,
            consumer_callback=self._profiler_wrapper(request_func),
            no_ack=True)

        return self  # Fluent pattern

    def add_topic(self, routing_key, request_func, queue=None, exclusive=False, exchange=amqppy.AMQP_EXCHANGE, durable=False,
                  auto_delete=True, no_ack=True, **kwargs):
        """ Register a new consumer task. These tasks will be execute when a new message arrive
        at :param:`queue_name`, which is binded to the exchange with :param:`routing_key`.
        """
        logger.debug("adding topic, exchange: {}, topic: {} --> {}".format(exchange, routing_key, request_func, kwargs))
        self.no_ack = no_ack
        channel = self._create_channel(exchange, request_func)
        queue_name = queue if queue else routing_key
        channel.queue_declare(queue=queue_name, exclusive=exclusive, durable=durable, auto_delete=auto_delete,
                              arguments=kwargs)
        channel.queue_bind(queue=queue_name, exchange=exchange, routing_key=routing_key)
        channel.basic_consume(
            queue=queue_name,
            consumer_callback=self._profiler_wrapper_topic(request_func),
            no_ack=no_ack)
        return self  # Fluent pattern

    def _profiler_wrapper(self, request_func):
        @wraps(request_func)
        def _wrapper(*args, **kwargs):
            logger.debug("request \'{}\'.*args: {}".format(request_func.__name__, args))
            # process request arguments
            deliver = args[1]
            properties = args[2]
            message = args[3]
            logger.debug("Starting request \'{}\'".format(request_func.__name__))
            # response = request_func(*args, **kwargs)
            start = time.time()
            try:
                response = {
                    # message is text, it should be converted in dictionary at request func
                    "result": request_func(exchange=deliver.exchange, routing_key=deliver.routing_key, headers=properties.headers, body=message),
                    "success": True
                }
            except Exception as e:
                logger.warning("Exception in request \'{}\', routing_key: {}\n{}".format(request_func.__name__,
                                                                                         deliver.routing_key,
                                                                                         traceback.format_exc()))
                response = {
                    "success": False,
                    "error": unicode(e)
                }
            elapsed = time.time() - start
            logger.debug('Request \'{}\' finished. Time elapsed: {}'.format(request_func.__name__, elapsed))

            # sending response back
            channel = self._callbacks[request_func].channel
            exchange = self._callbacks[request_func].exchange
            routing_key = properties.reply_to
            logger.debug('Sending RPC response to routing key: {}'.format(routing_key))
            try:
                publish_result = channel.basic_publish(
                    exchange=exchange,
                    routing_key=routing_key,
                    properties=pika.BasicProperties(
                        correlation_id=properties.correlation_id,
                        content_type='application/json'),
                    body=json.dumps(response),
                    mandatory=True)
                if not publish_result:
                    raise amqppy.PublishNotRouted("Request response was not routed")
            except amqppy.PublishNotRouted:
                # don't raise it
                logger.warning("RPC response it has not been published, it might be due to a response waiting timeout")
            except Exception:
                logger.error('Exception on publish message to routing_key: {}. Exception message: {}'.format(
                    routing_key, traceback.format_exc()))
            logger.debug('RPC response sended.')
        return _wrapper

    def _profiler_wrapper_topic(self, request_func):
        @wraps(request_func)
        def _wrapper(*args, **kwargs):
            logger.debug("topic \'{}\'.*args: {}".format(request_func.__name__, args))
            # logger.debug("request \'{}\'.**kwargs: {}".format(request_func.__name__, kwargs))
            # process request arguments
            deliver = args[1]
            properties = args[2]
            message = args[3]
            # logger.debug("Properties vars: {}".format(vars(properties)))
            logger.debug("Starting request \'{}\'".format(request_func.__name__))
            start = time.time()
            try:
                request_func(exchange=deliver.exchange, routing_key=deliver.routing_key, headers=properties.headers, body=message)
                if not self.no_ack:
                    self._callbacks[request_func].channel.basic_ack(delivery_tag=deliver.delivery_tag)
                    logger.debug("ACK sent")
            except amqppy.AbortConsume as e:
                logger.warning("AbortConsume exception: {}".format(e))
            except amqppy.DeadLetterMessage as e:
                logger.warning("DeadLetterMessage exception: {}".format(e))
                self._callbacks[request_func].channel.basic_reject(delivery_tag=deliver.delivery_tag, requeue=False)
            finally:
                elapsed = time.time() - start
                logger.debug('Request \'{}\' finished. Time elapsed: {}'.format(request_func.__name__, elapsed))

        return _wrapper

    def run(self):
        """ Start consumption """
        logger.debug('Running worker, waiting for the first message...')
        while not self.quit:
            self._conn.process_data_events()
            time.sleep(0.1)
        logger.debug("Exiting worker.")

    def run_async(self):
        self.thread = threading.Thread(target=self.run)
        self.thread.start()
        return self  # Fluent pattern

    def join(self):
        if self.thread:
            self.thread.join()



from setuptools import setup


long_description = ('amqppy is a very simplified AMQP client stacked over Pika. '
                    'amqppy is tested with RabbitMQ, but should also work '
                    'with other AMQP 0-9-1 brokers.')

setup(name='amqppy',
      version='0.0.6',
      description='amqppy is a very simplified AMQP client stacked over Pika',
      keywords=['amqp', 'client', 'rabbitmq', 'amqp client'],
      long_description=open('README.rst').read(),
      author='Marcel Janer Font',
      author_email='marceljanerfont@gmail.com',
      maintainer='Marcel Janer Font',
      maintainer_email='marceljanerfont@gmail.com',
      url='https://github.com/marceljanerfont/amqppy',
      packages=['amqppy'],
      license='MIT',
      install_requires=['pika'],
      package_data={'': ['LICENSE', 'README.rst']},
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: Jython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Topic :: Communications',
          'Topic :: Internet',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Networking'],
      zip_safe=True)
